/* This file should define some subset of the following:

   PATH_LOADSEARCH	The default value of `load-path'.

   PATH_EXEC		The default value of `exec-directory' and `exec-path'.
			(exec-path also contains the value of whatever is in
			the PATH environment variable.)

   PATH_LOCK		The name of the directory that contains lock files
			with which we record what files are being modified in 
			Emacs.  This directory should be writable by everyone.
			If this is specified, the string must end with a slash!

   PATH_SUPERLOCK	The name of the file !!!SuperLock!!! in the lock 
			directory.  You probably should let this default...

   The file startup.el guesses at reasonable values for load-path, exec-path,
   and lock-directory.  If any of the above constants are defined when emacs
   is built, they override the guesses made in startup.el.

   There is no need to specify anything in this file unless there is something
   strange about your emacs installation.  See startup.el for a description of
   the heuristic used.
 */
