;;; load the elisp side of Energize

(load "energize/energize-init")
(load "energize/energize-mode")
(load "energize/energize-menus")
(load "energize/energize-shell")
(load "energize/backtrace-logging")

(provide 'energize)
