/* ColumnWidget -- a simple Composite widget that manages its children in
 * a column.
 */

#ifndef _ColumnWidget_h
#define _ColumnWidget_h

typedef struct _ColumnClassRec *ColumnWidgetClass;
typedef struct _ColumnRec *ColumnWidget;

extern WidgetClass columnWidgetClass;

#endif /* _ColumnWidget_h */
