/*
** Generated by X-Designer 
** Modified by fidget-widget, 18-Aug-92
*/

/*
**LIBS: -lXm -lXmt -lX11
*/

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleBG.h>




Widget create_breaklist_p_sheet( parent )
Widget parent;
{
	Display *display = XtDisplay ( parent );
	Widget children[3];      /* Children to manage */
	Arg al[64];           /* Arg List */
	register int ac = 0;      /* Arg Count */
	char from_s [256];    /* For font list conversion */
	XrmValue from_value, to_value; /* ditto */
	int fg, bg;           /* colour values for pixmaps */ 
	short h_spacing, v_spacing; /* default spacing for forms */ 
	XmString *list_items; /* For list items */
	int list_item;        /* Index for list_items */
	XmString xmstrings[10];    /* temporary storage for XmStrings */
	Widget widget1;
	Widget widget2;
	Widget widget3;
	Widget widget4;
	Widget widget5;
	Widget widget6;
	Widget widget7;
	Widget widget8;
	Widget widget9;

	XtSetArg(al[ac], XmNallowShellResize, TRUE); ac++;
	ac = 0;
	XtSetArg(al[ac], XmNshadowType, XmSHADOW_OUT); ac++;
	widget1 = XmCreateFrame ( parent, "breakpointPanel", al, ac );
	ac = 0;
	XtSetArg(al[ac], XmNautoUnmanage, FALSE); ac++;
	widget2 = XmCreateForm ( widget1, "panel", al, ac );
	ac = 0;
	XtSetArg(al[ac], XmNorientation, XmHORIZONTAL); ac++;
	widget3 = XmCreateRadioBox ( widget2, "nameStyle", al, ac );
	ac = 0;
	widget4 = XmCreateToggleButtonGadget ( widget3, "nameOnly", al, ac );
	widget5 = XmCreateToggleButtonGadget ( widget3, "prototype", al, ac );
	widget6 = XmCreateToggleButtonGadget ( widget3, "mangled", al, ac );
	XtSetArg(al[ac], XmNorientation, XmHORIZONTAL); ac++;
	widget7 = XmCreateRadioBox ( widget2, "listingStyle", al, ac );
	ac = 0;
	widget8 = XmCreateToggleButtonGadget ( widget7, "terse", al, ac );
	widget9 = XmCreateToggleButtonGadget ( widget7, "verbose", al, ac );

	XtSetArg(al[ac], XmNtopAttachment, XmATTACH_WIDGET); ac++;
	XtSetArg(al[ac], XmNtopOffset, 2); ac++;
	XtSetArg(al[ac], XmNtopWidget, widget7); ac++;
	XtSetArg(al[ac], XmNbottomAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNbottomOffset, 2); ac++;
	XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNleftOffset, 2); ac++;
	XtSetArg(al[ac], XmNrightAttachment, XmATTACH_NONE); ac++;
        XtSetValues ( widget3,al, ac );
	ac = 0;

	XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNtopOffset, 2); ac++;
	XtSetArg(al[ac], XmNbottomAttachment, XmATTACH_NONE); ac++;
	XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
	XtSetArg(al[ac], XmNleftOffset, 2); ac++;
	XtSetArg(al[ac], XmNrightAttachment, XmATTACH_NONE); ac++;
        XtSetValues ( widget7,al, ac );
	ac = 0;
	children[ac++] = widget4;
	children[ac++] = widget5;
	children[ac++] = widget6;
	XtManageChildren(children, ac);
	ac = 0;
	children[ac++] = widget8;
	children[ac++] = widget9;
	XtManageChildren(children, ac);
	ac = 0;
	children[ac++] = widget3;
	children[ac++] = widget7;
	XtManageChildren(children, ac);
	ac = 0;
	children[ac++] = widget2;
	XtManageChildren(children, ac);
	ac = 0;
return widget1;
}


