/*
** Generated by X-Designer 
** Modified by fidget-widget, 18-Aug-92
*/

/*
**LIBS: -lXm -lXmt -lX11
*/

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>




Widget create_debugger_p_sheet( parent )
Widget parent;
{
	Display *display = XtDisplay ( parent );
	Widget children[11];      /* Children to manage */
	Arg al[64];           /* Arg List */
	register int ac = 0;      /* Arg Count */
	char from_s [256];    /* For font list conversion */
	XrmValue from_value, to_value; /* ditto */
	int fg, bg;           /* colour values for pixmaps */ 
	short h_spacing, v_spacing; /* default spacing for forms */ 
	XmString *list_items; /* For list items */
	int list_item;        /* Index for list_items */
	XmString xmstrings[10];    /* temporary storage for XmStrings */
	Widget widget1;
	Widget widget2;
	Widget widget3;
	Widget widget4;
	Widget widget5;
	Widget widget6;
	Widget widget7;
	Widget widget8;
	Widget widget9;
	Widget widget10;
	Widget widget11;
	Widget widget12;
	Widget widget13;

	XtSetArg(al[ac], XmNallowShellResize, TRUE); ac++;
	ac = 0;
	XtSetArg(al[ac], XmNshadowType, XmSHADOW_OUT); ac++;
	widget1 = XmCreateFrame ( parent, "debuggerPanel", al, ac );
	ac = 0;
	XtSetArg(al[ac], XmNspacing, 0); ac++;
	XtSetArg(al[ac], XmNmarginWidth, 0); ac++;
	XtSetArg(al[ac], XmNmarginHeight, 0); ac++;
	XtSetArg(al[ac], XmNorientation, XmHORIZONTAL); ac++;
	widget2 = XmCreateRowColumn ( widget1, "panel", al, ac );
	ac = 0;
	widget3 = XmCreatePushButton ( widget2, "start", al, ac );
	widget4 = XmCreatePushButton ( widget2, "run", al, ac );
	widget5 = XmCreatePushButton ( widget2, "stop", al, ac );
	widget6 = XmCreatePushButton ( widget2, "continue", al, ac );
	widget7 = XmCreatePushButton ( widget2, "step", al, ac );
	widget8 = XmCreatePushButton ( widget2, "next", al, ac );
	widget9 = XmCreatePushButton ( widget2, "finish", al, ac );
	widget10 = XmCreatePushButton ( widget2, "backtrace", al, ac );
	widget11 = XmCreatePushButton ( widget2, "up", al, ac );
	widget12 = XmCreatePushButton ( widget2, "down", al, ac );
	widget13 = XmCreatePushButton ( widget2, "print", al, ac );
	children[ac++] = widget3;
	children[ac++] = widget4;
	children[ac++] = widget5;
	children[ac++] = widget6;
	children[ac++] = widget7;
	children[ac++] = widget8;
	children[ac++] = widget9;
	children[ac++] = widget10;
	children[ac++] = widget11;
	children[ac++] = widget12;
	children[ac++] = widget13;
	XtManageChildren(children, ac);
	ac = 0;
	children[ac++] = widget2;
	XtManageChildren(children, ac);
	ac = 0;
return widget1;
}


